//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* manip.cpp: manipulate data (select, average, rearrange, ...)           *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <iostream>
#include <algorithm>
#include <boost/format.hpp>
#include <gsl/gsl_sort.h>

#include "mystd.h"
#include "olm.h"
#include "expr.h"
#include "asi.h"
#include "manip.h"
#include "dualplot.h" // wg. edif.h
#include "edif.h" // wg. ListZ()
#include "xax_lex.h"
#include "index.h"

using namespace std;
using boost::format;

// local routines:
namespace NManip {
    void ScaRemoveConstantZ( COlo *f );
    int PtsExeRetEqui( const vector<double> x, uint nblock, double tol, 
                       IndexSet *Sel);
}


//***************************************************************************//
//* restrict forthcoming operations                                         *//
//***************************************************************************//

//! Restrict forthcoming operations
void NManip::ProtectScans( bool on )
{
    string sel;
    if( on )
        sel = sask( "Protect which spectra from forthcoming operations" );
    else
        sel = "-";

    NOlm::IterateEle eiter;
    CEle *e;
    while((e=eiter())) {
        e->P()->ProtectedScans = CList( sel, 0, e->nScan()-1 );
        e->P()->lDoc.push_back( "m/ " + sel );
    }
 }


//***************************************************************************//
//* manipulations on data points                                            *//
//***************************************************************************//

//! Delete (or retain) points according to list of indices.

void NManip::PtsSelect( string del_or_ret )
{
    string action;
    bool sel_del;
    if        ( del_or_ret=="d" ){
        action = "Delete";
        sel_del = true;
    } else if ( del_or_ret=="r" ){
        action = "Retain";
        sel_del = false;
    } else
        throw string( "FATAL: invalid del_or_ret" );

    static string pSel = "";
    pSel = wask( action + " which points", pSel );
    CList pLis( pSel );

    const COld *fin; COld fout;
    const CScan *Sin; CScan Sout;
	
    NOlm::IterateD fiter;
    NOlm::SelNew();
    while((fin=fiter())) {
        fout = *fin;
        fout.lDoc.push_back( "mp" + del_or_ret + " " + pSel );
        fout.VS.clear();

        for (uint j=0; j<fin->VS.size(); j++) {
            Sin = &(fin->VS[j]);
            Sout = *Sin;
            Sout.x.clear();
            Sout.y.clear();
            Sout.dy.clear();
            pLis.evaluate( 0, Sin->size()-1 );
            for (uint i=0; i<Sin->size(); i++) {
                if ( sel_del ^ pLis.contains(i) )
                    Sout.push_xyd(Sin->x[i], Sin->y[i], Sin->dy[i]);
            }
            if (Sout.size())
                fout.VS.push_back(Sout);
        }
        if (fout.VS.size())
            NOlm::OloAdd(&fout);
        else 
            throw string( "no data point left" );
    }
}

//! Bin points.

void NManip::PtsAvge()
{
    static string groups;
    IndexSet Groups;
    IndexSetIterator ISet;
    uint i, igi, igf, ng;
    double xg, yg;

    groups = index_ask("Start groups at points", groups);
    if (groups=="") return;
	
    const COld *fin; COld fout;
    const CScan *Sin; CScan Sout;

    NOlm::IterateD fiter;
    NOlm::SelNew();
    while((fin=fiter())) {
        fout = *fin;
        fout.lDoc.push_back("mpa " + groups);
        fout.VS.clear();

        for (uint j=0; j<fin->VS.size(); j++) {
            Sin = &(fin->VS[j]);
            Sout = *Sin;
            Sout.x.clear();
            Sout.y.clear();
            if (Groups.parse(groups, 0, Sin->size()))
                throw "File " + strg(fiter.SelNo()) + ", scan " + strg(j) +
                    " invalid group";
            if (!Groups.contains(0))
                throw string( "Group must start at 0" );
            Groups.insert(Sin->size());

            ISet.reset(Groups);
            while (ISet(&igi) && ISet(&igf)) {
                // printf("DEBUG igi=%u igf=%u\n", igi, igf);
                ISet.stepback();
                xg = yg = 0;
                for (i=igi; i<igf; ++i) {
                    xg += Sin->x[i];
                    yg += Sin->y[i];
                }
                ng = igf - igi;
                Sout.push_xy(xg/ng, yg/ng);
            }
            fout.VS.push_back(Sout);
        }
        NOlm::OloAdd(&fout);
    }
}

//! Break one scan into several scans.

void NManip::PtsBreak()
{
    // determine scan length:
    const COld *fin; COld fout;
    NOlm::IterateD fiter;
	
    // ask for break points:
    static string pSel = "";
    pSel = wask( "Subscans start at entry numbers", pSel );
    CList pLis( pSel );

    // loop over input files:
    const CScan *Sin; CScan Sout;
    fiter.Reset();
    NOlm::SelNew();
    while((fin=fiter())) {
				
        // new z, documentation:
        fout = *fin;
        fout.VS.clear();
        fout.ZCo.push_back(CCoord(fin->xco.name + "_break", fin->xco.unit));
        fout.lDoc.push_back( "mpb " + pSel );
		
        // execute per scan:
        for(uint j=0; j<fin->nScan(); ++j) {

            uint np = fin->nPts(j);
            pLis.evaluate( 0, np-1 );

            Sin = &(fin->VS[j]);
			
            uint i0, ib, ii;
            // OBSOLETE: this is the last usage of ListIter.
            // eliminate as soon as possible
            ListIter piter(pLis);
            piter(&i0);
            do {
                if (!(ii = piter(&ib)))
                    ib = np;
                // new scan from i0 to ib-1:
                //printf("DEB subscan from %d to %d-1\n",i0,ib);
                if (ib<i0)
                    throw "invalid break "+strg(ib)+" < "+strg(i0);
                Sout.Clear();
                for (uint i=i0; i<ib; ++i)
                    Sout.push_xy(Sin->x[i], Sin->y[i]);
                Sout.z = Sin->z;
                Sout.z.push_back(Sout.x[0]);
                fout.VS.push_back(Sout);
                i0 = ib;
            } while (ii);
        }
        NOlm::OloAdd(&fout);
    }
}

//! Sort scan according to given expression.

void NManip::PtsSort()
{
    string expr = sask("Sort points according to");
    if (expr=="") return;
    PTree T;
    user_xaxparse( expr.c_str(), &T );

    NOlm::IterateD fiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    const COld *fin; COld fout;
    const COlo *f;
    const CScan *Sin; CScan Sout;
    uint k, n, i;
    while((fin=fiter())) {
        k = fiter.SelNo();
        f = (const COlo*) fin;
        fout = *fin;
        fout.VS.clear();
        fout.lDoc.push_back("mpo "+expr);

        for (uint j=0; j<fin->nScan(); j++) {
            Sin = &(fin->VS[j]);
            n = Sin->size();
            vector<double> v(n);
            T->tree_vec_val( &v, k, j );
            size_t P[n];
            gsl_sort_index (P, &(v[0]), 1, n);
            Sout = *Sin;
            vector<double> xout(n);
            for (i=0; i<n; ++i) {
                xout[i] = Sin->x[P[i]];
                Sout.y[i]  = Sin->y[P[i]];
                Sout.dy[i] = Sin->dy[P[i]];
            }
            Sout.x = xout;
            fout.VS.push_back(Sout);
        }
        NOlm::OloAdd(&fout);
    }
}

//! Average points that have the same x.

void NManip::PtsAvgeEq(void)
{
    uint j, i, ii, n, ng;
    double xg, yg;

    const COld *fin; COld fout;
    const CScan *Sin; CScan Sout;

    NOlm::IterateD fiter;
    NOlm::SelNew();
    while((fin=fiter())) {
        fout = *fin;
        fout.lDoc.push_back("mpq # average points when x equal");
        fout.VS.clear();

        for (j=0; j<fin->VS.size(); j++) {
            Sin = &(fin->VS[j]);
            n = Sin->size();
            Sout = *Sin;
            Sout.x.clear();
            Sout.y.clear();

            for (i=0; i<n; i=ii) {
                xg = Sin->x[i];
                yg = Sin->y[i];
                ng = 1;
                for (ii=i+1; ii<n; ++ii) {
                    if (Sin->x[ii] != xg) break;
                    yg += Sin->y[ii];
                    ++ng;
                }
                Sout.push_xy(xg, yg/ng);
            }
            fout.VS.push_back(Sout);
        }
        NOlm::OloAdd(&fout);
    }
}

//! Delete points unless x is equidistant. Obsolete (use case forgotten).

void NManip::PtsRetainEqui()
{
    const COld *fin; COld fout;
    const CScan *Sin; CScan Sout;
    vector<double> x;
    uint n, i, ii, offs;
    static uint nblock=1;
    double step, tol=1e-6;

    nblock = iask( "Block size", nblock );
    if ( nblock==0 ) return;

    NOlm::IterateD fiter;
    NOlm::SelNew();
    while( (fin=fiter()) ) {
        fout = *fin;
        fout.lDoc.push_back("mpdb # delete points unless x is equidistant...");
        fout.VS.clear();

        for (uint j=0; j<fin->VS.size(); j++) {
            Sin = &(fin->VS[j]);
            Sout = *Sin;
            x = Sin->x;
            n = x.size();
            if ( n<2 )
                throw string("not enough data points" );
            // determine step :
            step = x[1] - x[0];
            // retain blocks:
            offs=0;
            for (i=0; i<=n-nblock; ) {
                for (ii=i+1; ii<i+nblock; ii++) {
                    if (fabs(x[ii]-x[ii-1]-step) > tol) {
                        printf("unselect %u to %u\n", 
                               i, ii-1);
                        Sout.erase(offs+i,offs+ii);
                        offs -= ii-i;
                        break;
                    } 
                }
                i = ii;
            }
            if (i<n) printf("overhang %u to %u\n", i, n-1);
            Sout.erase(offs+i, offs+n);
            if (Sout.size())
                fout.VS.push_back(Sout);
        }
        if (fout.VS.size())
            NOlm::OloAdd(&fout);
        else 
            throw string( "no data point left" );
    }
}

//! Obsolete (too special, ill documented).

void NManip::PtsGroupWeighted()
// for use in PISA jan07
// input: each scan's y tuple contains: xg, wgt, y0, .., ym
//    where xg is a coordinate according to which the y0 .. ym will be grouped
//    where wgt is the weight of the tuple
{
    // cout << "warning:\n"
    //    "will fail if input point ought to be distributed"
    //    " in more than 2 groups\n";
    static int ng=40;
    ng = iask( "number of groups", ng );
    if( ng<2 ) return;

    NOlm::IterateD fiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    const COld *fin;
    COld fout, fout2;
    const COlo *f;
    const CScan *Sin;
    CScan Sout, Sout2;
    double wgttot, wgtgrp, wgtacc, frac;
    uint k, nz, n, ny, i, iy, igrp;
    vector<double> xtmp, wtmp;

    while((fin=fiter())) {
        k = fiter.SelNo();
        nz = fin->ZCo.size();
        f = (const COlo*) fin;

        ny = fin->VS[0].size()-2;
        if( (int)ny<1 ){
            printf( "ny<1\n" );
            return;
        }

        fout = *fin;
        fout.VS.clear();

        fout.xco = CCoord( "group", "" );
        fout.ZCo.pop_back();
        if( ny>1 ) fout.ZCo.push_back( CCoord( "iy", "" ) );

        fout2 = fout;
        fout2.yco.name = "stdv["+fout.yco.name+"]";
        fout.lDoc.push_back("mpgw "+strg(ng));
        fout2.lDoc.push_back("error bars of mpgw "+strg(ng));

        Sout.Clear();
        Sout2.Clear();

        vector<double> xgrp(ng);
        vector< vector<double> > ytmp, ygrp, ygrp2;
        ytmp.resize( ny );
        ygrp.resize( ny );
        ygrp2.resize( ny );
        for( iy=0; iy<ny; ++iy ){
            ytmp [iy].clear();
            ygrp [iy].resize(ng);
            ygrp2[iy].resize(ng);
        }

        for (uint j=0; j<fin->nScan(); j++) {

            Sin = &(fin->VS[j]);
            if( Sin->size()-2 != ny )
                throw string( "ny not constant" );
            Sout.z = Sin->z;
            Sout.z.pop_back();
            Sout2.z = Sin->z;
            Sout2.z.pop_back();
            xtmp.push_back( Sin->y[0] );
            wtmp.push_back( Sin->y[1] );
            for( iy=0; iy<ny; ++iy )
                ytmp[iy].push_back( Sin->y[2+iy] );

            if ( j+1==fin->VS.size() || 
                 nz>=2 && fin->VS[j+1].z[nz-2]!=Sin->z[nz-2] ) {
                // reached end-of-file or end-of-relevant-z-group
                // sort, group, save, possibly start new scan
                n = xtmp.size();
                size_t P[n];
                double xin[n];
                for (i=0; i<n; ++i) { xin[i] = xtmp[i]; }
                gsl_sort_index( P, xin, 1, n );
                vector<double> xsor(n), wsor(n);
                vector< vector<double> > ysor(ny);
                for( iy=0; iy<ny; ++iy ){
                    ysor[iy].resize(n);
                }
                for (i=0; i<n; ++i) {
                    xsor[i] = xtmp[P[i]];
                    wsor[i] = wtmp[P[i]];
                    for( iy=0; iy<ny; ++iy )
                        ysor[iy][i] = ytmp[iy][P[i]];
                }
                for( igrp=0; igrp<ng; ++igrp ){
                    xgrp[igrp] = 0;
                    for( iy=0; iy<ny; ++iy ){
                        ygrp[iy][igrp] = 0;
                        ygrp2[iy][igrp] = 0;
                    }
                }
                wgttot = 0;
                for( i=0; i<n; ++i ){ wgttot += wsor[i]; }
                wgtgrp = wgttot/ng;
                wgtacc = 0;
                for( i=0, igrp=0; i<n; ++i ){
                    if( wgtacc+wsor[i]>(igrp+1)*wgtgrp ){
                        frac = (igrp+1)*wgtgrp-wgtacc;
                        xgrp[igrp] += frac * xsor[i] / wgtgrp;
                        for( iy=0; iy<ny; ++iy ){
                            ygrp[iy][igrp] += frac * ysor[iy][i] / wgtgrp;
                            ygrp2[iy][igrp] += frac* SQR(ysor[iy][i]) /wgtgrp;
                        }
                        ++igrp;
                        if( igrp>=ng ) break;
                        frac = wgtacc+wsor[i]-igrp*wgtgrp;
                        xgrp[igrp] += frac * xsor[i] / wgtgrp;
                        for( iy=0; iy<ny; ++iy ){
                            ygrp[iy][igrp] += frac * ysor[iy][i] / wgtgrp;
                            ygrp2[iy][igrp] += frac* SQR(ysor[iy][i]) /wgtgrp;
                        }
                    } else {
                        xgrp[igrp] += wsor[i] * xsor[i] / wgtgrp;
                        for( iy=0; iy<ny; ++iy ){
                            ygrp[iy][igrp] += wsor[i] * ysor[iy][i] / wgtgrp;
                            ygrp2[iy][igrp] += wsor[i]*SQR(ysor[iy][i])/wgtgrp;
                        }
                    }
                    wgtacc += wsor[i]; 
                }
                for( iy=0; iy<ny; ++iy ){
                    if( ny>1 ){
                        Sout.z.push_back( iy );
                        Sout2.z.push_back( iy );
                    }
                    for( igrp=0; igrp<ng; ++igrp ){
                        Sout.push_xy( xgrp[igrp], ygrp[iy][igrp] );
                        Sout2.push_xy( xgrp[igrp], 
                            sqrt( ygrp2[iy][igrp] - SQR(ygrp[iy][igrp]) ) );
                    }
                    fout.VS.push_back(Sout);
                    Sout.Clear();
                    Sout.z = Sin->z;
                    Sout.z.pop_back();
                    fout2.VS.push_back(Sout2);
                    Sout2.Clear();
                    Sout2.z = Sin->z;
                    Sout2.z.pop_back();
                }
                xtmp.clear();
                wtmp.clear();
                for( iy=0; iy<ny; ++iy )
                    ytmp[iy].clear();
            }
        }
        NOlm::OloAdd(&fout);
        NOlm::OloAdd(&fout2);
    }
}

//! Symmetrize scan.

void NManip::PtsSymmetrize()
{
    static string expr = "0";
    expr = sask("Assume symmetry with mirror positioned at", expr);
    if (expr=="") return;
    PTree T;
    user_xaxparse( expr.c_str(), &T );

    NOlm::IterateD fiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    const COld *fin; COld fout;
    const COlo *f;
    const CScan *sin; CScan Sout;
    uint k, n, i, il, ih, nl, nh;
    while((fin=fiter())) {
        k = fiter.SelNo();
        f = (const COlo*) fin;
        fout = *fin;
        fout.VS.clear();
        fout.lDoc.push_back("mpsym "+expr);

        for (uint j=0; j<fin->nScan(); j++) {
            sin = &(fin->VS[j]);
            Sout.Clear();
            if( mystd::sorted( sin->x )!=1 )
                throw string( "not sorted" );
            double step;
            if( !mystd::is_equidist( &step, sin->x ) )
                throw string( "not equidistant" );
            double centre;
            T->tree_point_val( &centre, k, j );
            il = (int) ( (centre - sin->x[0])/step );
            ih = il + 1;
            n = sin->size();
            nl = il + 1;
            nh = n - ih;
            if( nl<1 || nh<1 )
                throw string( "nothing to symmetrize" );
            for (i=0; i<(nl<nh?nl:nh); ++i) {
                Sout.push_xy( ( -sin->x[il-i] + sin->x[ih+i] ) / 2,
                                ( sin->y[il-i] + sin->y[ih+i] ) / 2 );
            }
            if( nl<nh ){
                for (i=nl; i<nh; ++i)
                    Sout.push_xy(  sin->x[ih+i], sin->y[ih+i] );
            } else {
                for (i=nh; i<nl; ++i)
                    Sout.push_xy( -sin->x[il-i], sin->y[il-i] );
            }
            Sout.z = sin->z;
            fout.VS.push_back(Sout);
        }
        NOlm::OloAdd(&fout);
    }
}

//! Delete error bars.

void NManip::PtsErrorDelete()
{
    NOlm::IterateD fiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    const COld *fin; COld fout;
    CScan Sout;
    while((fin=fiter())) {
        fout = *fin;
        fout.VS.clear();
        fout.lDoc.push_back("mdy-");
        for ( uint j=0; j<fin->nScan(); j++ ) {
            Sout = fin->VS[j];
            Sout.dy.clear();
            fout.VS.push_back(Sout);
        }
        NOlm::OloAdd(&fout);
    }
}


//***************************************************************************//
//* manipulations on scans                                                  *//
//***************************************************************************//

//! Remove z variable that does not vary with j.

void NManip::ScaRemoveConstantZ( COlo *f )
{
    uint nz = f->nZ();
    for( int iz=nz-1; iz>=0; --iz ){
        uint nj = f->nScan();
        if( nj<1 )
            throw string( "file contains no scan" );
        double z = f->z(0,iz);
        bool z_is_const = true;
        for( uint j=1; j<nj; ++j ){
            if( f->z(j,iz) != z ){
                z_is_const = false;
            }
        }
        if( z_is_const ){
            printf( " constant z%i becomes r%zu\n", iz, f->RPar.size() );
            f->RPar.push_back( CParam( f->ZCo[iz], z ) );
            f->ZCo.erase( f->ZCo.begin()+iz );
            f->erase_z( iz );
        }
    }
}

//! Delete (or retain) scans according to list of indices.

void NManip::ScaSelect( string del_or_ret )
{
    string action;
    bool sel_ret;
    if        ( del_or_ret=="d" ){
        action = "Delete";
        sel_ret = false;
    } else if ( del_or_ret=="r" ){
        action = "Retain";
        sel_ret = true;
    } else
        throw string( "BUG: invalid del_or_ret" );

    CList JSel, JSelSorted;
    static string jSel = "";
    NOlm::JSelAsk( action + " which scans", &jSel, &JSel );

    const CEle *ein;
    NOlm::IterateEle eiter;
    NOlm::SelNew();
    while((ein=eiter())) {
        COlo *fout;
        fout = ein->copy();
        JSel.evaluate( 0, ein->nScan()-1 );
        JSelSorted = JSel;
        JSelSorted.SortReverse();
        fout->lDoc.push_back( "ms" + del_or_ret + " " + JSel.str());
        for( int j=ein->nScan()-1; j>=0; --j )
            if( sel_ret ^ JSelSorted.contains( j ) )
                fout->erase( j );
        ScaRemoveConstantZ( fout );
        NOlm::OloAdd( fout );
        delete fout;
    }
}

void NManip::ScaAvge()
{
    CList JSel;
    static string jSel = "";
    NOlm::JSelAsk( "Start groups at scans", &jSel, &JSel );

    const CEle *ein;
    NOlm::IterateEle eiter;
    NOlm::SelNew();
    while( (ein=eiter()) ) {
        COld* din;
        COlc* cin;
        din = ein->D();
        cin = ein->C();
        COlo *fout;
        fout = ein->copy();
        JSel.evaluate( 0, ein->nScan()-1 );
        if (JSel.V.size()<1 || JSel.V[0]!=0)
            throw string( "scan selection must contain 0" );
        fout->lDoc.push_back("msa " + JSel.str());
        fout->clear_V();
        for(uint iv=0; iv<JSel.size(); iv++) {
            uint ji = JSel.V[iv];
            uint jf = ( iv<JSel.size()-1 ? JSel.V[iv+1] : ein->nScan() );
            uint nz = ein->P()->nZ();
            vector<double> zout(nz);
            for( uint iz=0; iz<nz; ++iz ){
                uint mz = 0;
                double z = 0;
                for ( uint jj=ji; jj<jf; jj++ ) {
                    z += ein->P()->z(jj,iz);
                    ++mz;
                }
                zout[iz] = z/mz;
            }
            if( din ) {
                CScan Sout = din->VS[ji];
                uint n = din->VS[ji].size();
                for ( uint jj=ji+1; jj<jf; jj++){
                    if ( din->VS[jj].size() != din->VS[ji].size() )
                        throw "scan " + strg(jj) +
                            " has other size than scan " + strg(ji);
                    for( uint i=0; i<din->VS[jj].size(); ++i )
                        if ( din->VS[jj].x[i] != din->VS[ji].x[i] )
                            throw "scan " + strg(jj) + " has other x[" +
                                strg(i) + "] than scan 0";
                }
                vector<double> y(n);
                for ( uint i=0; i<n; i++ ){
                    uint mj = 0;
                    double ym = 0;
                    for ( uint jj=ji; jj<jf; jj++) {
                        ym += din->VS[jj].y[i];
                        ++mj;
                    }
                    y[i] = mj==0 ? 0 : ym/mj;
                }
                Sout.y = y;
                Sout.z = zout;
                ((COld*)fout)->VS.push_back(Sout);
            } else if ( cin ) {
                CCurve Cout;
                uint np = cin->nPar();
                vector<double> p(np);
                for ( uint ip=0; ip<np; ip++ ){
                    uint mj = 0;
                    double pm = 0;
                    for ( uint jj=ji; jj<jf; jj++) {
                        pm += cin->VC[jj].P[ip];
                        ++mj;
                    }
                    p[ip] = mj==0 ? 0 : pm/mj;
                }
                Cout.P = p;
                Cout.z = zout;
                ((COlc*)fout)->VC.push_back(Cout);
            }
        }
        ScaRemoveConstantZ( fout );
        NOlm::OloAdd( fout );
        delete fout;
    }
}

//! Merge scans, concatenating points.

void NManip::ScaJoin()
{
    CList JSel;
    static string jSel = "";
    NOlm::JSelAsk( "Start groups at scans", &jSel, &JSel );
    const COld *fin; COld fout;
    CScan Sout;
    vector<double> y;
    uint ji, jf, n, nz;
    string err;
    NOlm::IterateD fiter;
    NOlm::SelNew();
    while((fin=fiter())) {
        fout = *fin;
        JSel.evaluate( 0, fin->nScan()-1 );
        if (JSel.V.size()<1 || JSel.V[0]!=0)
            throw string( "scan selection must contain 0" );
        fout.lDoc.push_back("msj " + JSel.str());
        fout.VS.clear();
        nz = fin->nZ();
        for(uint iv=0; iv<JSel.size(); iv++) {
            ji = JSel.V[iv];
            jf = ( iv<JSel.size()-1 ? JSel.V[iv+1] : fin->nScan() );
            Sout = fin->VS[ji];
            n = fin->VS[ji].size();
            for ( uint jj=ji+1; jj<jf; jj++){
                for( uint i=0; i<fin->VS[jj].size(); ++i )
                    Sout.push_xy( fin->VS[jj].x[i], fin->VS[jj].y[i] );
            }
            fout.VS.push_back(Sout);
        }
        ScaRemoveConstantZ( &fout );
        NOlm::OloAdd(&fout);
    }
}

//! After each span, insert given number of copies.

void NManip::ScaSpawn()
{
    static int njj = 2;
    int njjIn = iask("output will have how many copies of input scan", njj);
    if (njjIn<1) return;
    njj = njjIn;

    const CEle *e;
    NOlm::IterateEle iter;
    NOlm::SelNew();
    while( (e=iter()) ){
        COlo *fout;
        fout = e->copy();
        fout->clear_V();
        fout->ZCo.push_back(CCoord("no-in-spawn", ""));
        fout->lDoc.push_back( "ms* " + strg(njj) );

        for( uint jj=0; jj<njj; ++jj ){
            for( uint j=0; j<e->nScan(); j++ ){
                if( e->isD() ){
                    CScan Sout = e->D()->VS[j];
                    Sout.z.push_back(jj);
                    ((COld*)fout)->VS.push_back(Sout);
                } else {
                    CCurve Sout = e->C()->VC[j];
                    Sout.z.push_back(jj);
                    ((COlc*)fout)->VC.push_back(Sout);
                }
            }
        }
        NOlm::OloAdd(fout);
        delete fout;
    }
}

//! Exchange x and z.

void NManip::ScaExch()
{
    const COld *fin; COld fout;
    CScan Sout;
    vector<double> zcommon, xcommon;
    uint ji, jf, j, i, ii, izco;
    CCoord xcoin, zcoin;
    string doc;

    NOlm::IterateD fiter;
    NOlm::SelNew();
    while((fin=fiter())) {
        fout = *fin;
        fout.VS.clear();

        xcoin = fin->xco;
        izco = fin->ZCo.size() - 1;
        if ( izco==(uint)-1 )
            throw string( "no input z coordinate" );
        zcoin = fin->ZCo[izco];

        doc = "msx # exchanging x=" + xcoin.str() + " against z" + strg(izco)
            + "=" + zcoin.str();
        cout << doc+"\n";
	
        fout.lDoc.push_back(doc);
        fout.xco = zcoin;
        fout.ZCo[izco] = xcoin;
		
        if ( fin->VS.size()<=0 )
            throw string( "no scans in file" );

        for (ji=0; ji<fin->VS.size(); ) {

            zcommon.clear();
            for (i=0; i<izco; ++i)
                zcommon.push_back(fin->VS[ji].z[i]);

            for (jf=ji+1; jf<fin->VS.size(); jf++)
                for (i=0; i<izco; ++i)
                    if (fin->VS[jf].z[i] != zcommon[i])
                        goto end_group;
        end_group:

            // build common grid:
            xcommon.clear();
            for (j=ji; j<jf; ++j)
                for (i=0; i<fin->VS[j].size(); ++i)
                    xcommon.push_back(fin->VS[j].x[i]);
            sort(xcommon.begin(), xcommon.end());
            mystd::unique(&xcommon, 1e-180, 1e-20);

            printf("group of scans %u .. %u has %zu different x\n", 
                   ji, jf-1, xcommon.size());

            ii = 0; // a guess
            for(i=0; i<xcommon.size(); ++i) {
                Sout.Clear();
                Sout.z = zcommon;
                Sout.z.push_back(xcommon[i]);

                for (j=ji; j<jf; ++j){
                    if( fin->VS[j].x[ii]!=xcommon[i] ){ // guess failed
                        for (ii=0; ii<fin->VS[j].size(); ++ii)
                            if( fin->VS[j].x[ii]==xcommon[i] )
                                break;
                    }
                    if( ii<fin->VS[j].size() )
                        Sout.push_xy( fin->VS[j].z[izco], fin->VS[j].y[ii] );
                }
                fout.VS.push_back(Sout);
                ++ii;
            }
            ji = jf;
        }
        NOlm::OloAdd(&fout);
    }
}

//! Sort scans according to expression.

void NManip::ScaSortByExpr()
{
    string expr = sask("Sort spectra according to");
    if (expr=="") return;
    PTree T;
    user_xaxparse( expr.c_str(), &T );

    NOlm::IterateD fiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    const COld *fin; COld fout;
    const COlo *f;
    const CScan *Sin; CScan Sout;
    uint k, nj;
    while((fin=fiter())) {
        k = fiter.SelNo();
        f = (const COlo*) fin;

        fout = *fin;
        fout.VS.clear();
        fout.lDoc.push_back("mso "+expr);

        nj = fin->nScan();
        vector<double> v(nj);
        size_t P[nj];
        for (uint j=0; j<nj; j++) {
            Sin = &(fin->VS[j]);
            T->tree_point_val( &(v[j]), k, j );
        }
        
        gsl_sort_index (P, &(v[0]), 1, nj);
                
        for (uint j=0; j<nj; j++) {
            fout.VS.push_back( fin->VS[P[j]] );
        }
        NOlm::OloAdd(&fout);
    }
}

//! Sort scans according to full z vector.

void NManip::ScaSortByZ()
{
    const COld *fin; COld fout;
    NOlm::IterateD fiter;
    NOlm::SelNew();
    while((fin=fiter())) {
        fout = *fin;
        sort( fout.VS.begin(), fout.VS.end(), CompareZ );
        NOlm::OloAdd(&fout);
    }
}

//! Change order of z coordinates.

void NManip::ZExchange()
{
    int nzmax = 0;
    int nzmin;
    char mod;
    string com;
    int num1;
    int num2;
    const COld *fin; COld fout;
    NOlm::IterateD fiter;
    while( (fin=fiter()) )
        nzmax = max( (int)nzmax, (int)fin->ZCo.size() );
    if       ( nzmax<=1 )
        throw string( "nothing to exchange" );
    else if( nzmax==2 ){
        mod = 'r';
        printf( "exchanging two z coordinates\n" );
        com = "exchanged two z coordinates";
        num1 = 1;
        nzmin = 2;
    } else {
        string sel;
        sel = wask( "rotate(r,rr,...), revert(m), exchange(<number>)" );
        if      ( sel[0]=='r' ){
            mod = 'r';
            num1 = 1;
            while( sel.length() > num1 ){
                if( sel[num1++]!='r' )
                    throw string( "invalid selection" );
            }
            com = "rotated z coordinates by " + strg(num1);
            nzmin = 2;
        } else if( sel=="m" ){
            mod = 'm';
            com = "reverted z coordinates";
            nzmin = 2;
        } else {
            if( sscanf(sel.c_str(), "%d", &num1)!=1 )
                throw string( "invalid selection" );
            if( num1>=nzmax )
                throw "no more than " + strg(nzmax) + " z coordinates";
            num2 = iask( "exchange with coordinate" );
            if( num2>=nzmax )
                throw "no more than " + strg(nzmax) + " z coordinates";
            else if( num1==num1 )
                throw string( "nothing to exchange" );
            mod = 'x';
            com = "exchanged z coordinates "+strg(num1)+" and "+strg(num2);
            nzmin = max( num1, num2 )+1;
        }
    }
                
    NOlm::SelNew();
    fiter.Reset();
    while((fin=fiter())) {
        fout = *fin;
        fout.lDoc.push_back( com );
        int nz = fout.ZCo.size();
        if      ( nz<nzmin ){
            printf( "WARNING nz<nzmin\n" );
        } else if( mod=='r' ){
            for( int irot=0; irot<num1; ++irot ){
                for( int iz=1; iz<nz; ++iz ){
                    fout.ZCo[iz] = fin->ZCo[iz-1];
                }
                fout.ZCo[0] = fin->ZCo[nz-1];
            }
        } else if( mod=='m' ){
            for( int iz=0; iz<nz; ++iz ){
                fout.ZCo[iz] = fin->ZCo[nz-1-iz];
            }
        } else { // mod=='x'
            fout.ZCo[num1] = fin->ZCo[num2];
            fout.ZCo[num2] = fin->ZCo[num1];
        }

        for (uint j=0; j<fin->VS.size(); j++) {
            if      ( nz<nzmin ){
                // do nothing
            } else if( mod=='r' ){
                for( int irot=0; irot<num1; ++irot ){
                    for( int iz=1; iz<nz; ++iz ){
                        fout.VS[j].z[iz] = fin->VS[j].z[iz-1];
                    }
                    fout.VS[j].z[0] = fin->VS[j].z[nz-1];
                }
            } else if( mod=='m' ){
                for( int iz=0; iz<nz; ++iz ){
                    fout.VS[j].z[iz] = fin->VS[j].z[nz-1-iz];
                }
            } else  { // mod=='x'
                fout.VS[j].z[num1] = fin->VS[j].z[num2];
                fout.VS[j].z[num2] = fin->VS[j].z[num1];
            }
        }
        NOlm::OloAdd(&fout);
    }

}

//! Delete z coordinate.

void NManip::ZDelete()
{
    string sel;
    NEdif::ListZ();
    sel = wask("Delete z coordinates (-=quit)");
    if (sel=="-" || sel=="q") return;
	
    const COld *fin; COld fout;

    NOlm::IterateD fiter;
    IndexSet ISel;
    IndexSetIterator II;
    uint iz;

    NOlm::SelNew();
    while((fin=fiter())) {
        fout = *fin;
        fout.lDoc.push_back("mz- "+sel);

        if ( ISel.parse(sel, 0, fin->ZCo.size()) )
            throw ( "invalid selection for file " + strg( fiter.SelNo() ) );

        II.setback(ISel);
        while (II(&iz)) {
            fout.ZCo.erase(fout.ZCo.begin()+iz);
            for (uint j=0; j<fin->VS.size(); j++) 
                fout.VS[j].z.erase(fout.VS[j].z.begin()+iz);
        }

        NOlm::OloAdd(&fout);
    }
}

//! Break file into several files, according to z0.

void NManip::ScaBreak()
{
    const COld *fin;
    COld ftmp, fout;
    CCoord zco;
    double zval;
    NOlm::IterateD fiter;
    NOlm::SelNew();
    while((fin=fiter())) {
        if (!(fin->ZCo.size()))
            throw string( "invalid operation / no z coordinate" );
        zco = fin->ZCo[0];
        cout << ".. eliminate " << zco << "\n";

        ftmp = *fin;
        ftmp.lDoc.push_back("scans -> files, eliminating " + zco.str());

        ftmp.ZCo.erase(ftmp.ZCo.begin());

        for (uint j=0; j<ftmp.VS.size(); ) {
            fout = ftmp;
            fout.VS.clear();
            zval = ftmp.VS[j].z[0];
            fout.RPar.push_back( CParam( zco, zval ) );
            ftmp.VS[j].z.erase(ftmp.VS[j].z.begin());
            fout.VS.push_back(ftmp.VS[j]);
            while (++j<ftmp.VS.size() && ftmp.VS[j].z[0]==zval) {
                ftmp.VS[j].z.erase(ftmp.VS[j].z.begin());
                fout.VS.push_back(ftmp.VS[j]);
            }
            NOlm::OloAdd(&fout);
        }
    }
}


//***************************************************************************//
//* manipulations on files                                                  *//
//***************************************************************************//

//! Merge files, concatenating scans.

void NManip::FilMerge( const string& opts )
{
    const COld *fin; COld fout;
    CCoord zco;
    vector<CParam> RPar;    
    vector<bool> RParDiff;    
    NOlm::IterateD fiter;
    NOlm::SelNew();
    if (!(fin=fiter()))
        throw string( "no input files given" );

    /** parse options **/
    bool add_zk = false;
    if      ( opts=="" )
        ;
    else if ( opts=="+" )
        add_zk = true;

    /** coordinates **/
    CCoord co;
    fiter.Reset();
    co = fiter()->xco;
    while( (fin=fiter()) )
        if ( fin->xco != co )
            throw string( "different x coordinates" );
    fout.xco = co;
    fiter.Reset();
    co = fiter()->yco;
    while ((fin=fiter()))
        if (fin->yco != co)
            throw string( "different y coordinates" );
    fout.yco = co;
    vector<CCoord> ZCo;
    fiter.Reset();
    ZCo = fiter()->ZCo;
    while ((fin=fiter())) {
        if (fin->ZCo.size() != ZCo.size())
            throw string( "different numer of z coordinates" );
        for (uint i=0; i<ZCo.size(); ++i)
            if (fin->ZCo[i] != ZCo[i])
                throw "different z"+strg(i)+" coordinate";
    }
    fout.ZCo = ZCo;

    /** new z coordinate **/
    if ( add_zk )
        fout.ZCo.push_back(CCoord("number-of-input-file", ""));

    /** check RPar **/
    fiter.Reset();
    RPar = fiter()->RPar;
    RParDiff.resize( RPar.size() );
    for (uint i=0; i<RPar.size(); ++i)
        RParDiff[i] = false;
    while ((fin=fiter())) {
        if (fin->RPar.size() != RPar.size())
            throw string( "different numer of r parameters");
        for (uint i=0; i<RPar.size(); ++i)
            if (fin->RPar[i].Co != RPar[i].Co )
                throw string( "different r parameter");
        for (uint i=0; i<RPar.size(); ++i)
            if (fin->RPar[i].val != RPar[i].val )
                RParDiff[i] = true;
    }

    for (uint i=0; i<RPar.size(); ++i)
        if( RParDiff[i] )
            fout.ZCo.push_back( RPar[i].Co );

    /** file name **/
    string fnam;
	
    fiter.Reset();
    fnam = fiter()->name;
    while ((fin=fiter())) {
        uint n2=fin->name.size();
        if (n2<fnam.size())
            fnam.erase(fnam.begin()+n2, fnam.end());
        for (uint i=0; i<fnam.size(); ++i)
            if (fnam[i]!=(fin->name)[i])
                fnam[i] = ' ';
    }
    for (string::iterator cp = fnam.begin(); cp != fnam.end(); ++cp)
        if (*cp==' ') fnam.erase(cp--);
	
    if (fnam == "") {
        static uint callNo=0;
        fnam = str( format( "merge%03u" ) % callNo++ );
    }

    fnam = wask("Joined file name", fnam);
    if (fnam=="")
        return;
    fout.name = fnam;

    /** doc lines **/
    fout.lDoc.push_back(fnam + " is merger of:");
    fiter.Reset();
    while ((fin=fiter())) {
        fout.lDoc.push_back("  " + fin->name);
        for (uint i=0; i<fin->lDoc.size(); ++i)
            fout.lDoc.push_back("    " + fin->lDoc[i]);
    }

    /** now merge the data **/
    CScan S;
    fiter.Reset();
    while ((fin=fiter())) {
        for (uint j=0; j<fin->nScan(); j++) {
            S = fin->VS[j];
            if ( add_zk )
                S.z.push_back(fiter.index());
            for (uint i=0; i<RPar.size(); ++i)
                if( RParDiff[i] )
                    S.z.push_back( fin->RPar[i].val );
            fout.VS.push_back(S);
        }
    }
    NOlm::OloAdd(&fout);
}

//! Merge files with same number of scans, concatenating scans with same j.

void NManip::FilMergePointwise()
{
    const COld *fin; COld fout;
    const CScan *Sin; CScan *Sout;
    uint i, j;
    NOlm::IterateD fiter;
    NOlm::SelNew();

    if ( !(fin=fiter()) )
         throw string("no input files given");

    fout = *fin;

    while ((fin=fiter())) {
        if ( fin->nScan() != fout.nScan() )
            throw string( "different # scans" );

        if ( fin->xco != fout.xco || fin->yco != fout.yco )
            throw string( "different x/y coordinates" );
		
        for (j=0; j<fout.nScan(); ++j) {
            Sin = &(fin->VS[j]); 
            Sout = &(fout.VS[j]); 
            if ( Sin->z.size() != Sout->z.size() )
                 throw string( "different # z entries" );
            for (i=0; i<Sout->z.size(); ++i)
                if ( Sin->z[i] != Sout->z[i] )
                    throw string( "different # z values" );
            for (i=0; i<Sin->size(); ++i)
                Sout->push_xy(Sin->x[i], Sin->y[i]);
        }

        fout.lDoc.push_back("appended pointwise " + fin->name);
    }

    NOlm::OloAdd(&fout);
}
